use std::{collections::HashMap, rc::Rc};

use crate::{
    DeBrujinAst,
    parse::{Constant, Ident},
    types::{PrimitiveType, TaggedType, Type},
    vec_map::VecMap,
};

use super::InferError;

/// Old algorithm use unification instead
pub fn infer_type(
    gamma: &HashMap<Ident, TaggedType>,
    ast: DeBrujinAst,
) -> Result<TaggedType, InferError> {
    infer_type_debrujin_int(gamma, Rc::new(VecMap::new()), ast)
}

fn infer_type_debrujin_int(
    gamma_free: &HashMap<Ident, TaggedType>,
    mut gamma_bound: Rc<VecMap<usize, TaggedType>>,
    ast: DeBrujinAst,
) -> Result<TaggedType, InferError> {
    match ast {
        DeBrujinAst::Abstraction(_, arg_type, ast) => {
            let gamma_ref = Rc::make_mut(&mut gamma_bound);
            gamma_ref.map_keys(|i| *i += 1);
            gamma_ref.insert(1, arg_type.clone().unwrap());

            let out_type = infer_type_debrujin_int(gamma_free, gamma_bound, *ast)?;
            // TODO: Fix this hack
            let typ = arg_type.unwrap().make_arrow(out_type);
            Ok(typ)
        }
        DeBrujinAst::Application(lhs, rhs) => {
            let left_type = infer_type_debrujin_int(gamma_free, gamma_bound.clone(), *lhs)?;
            let Some((in_type, out_type)) = left_type.split_arrow() else {
                return Err(InferError::NotAFunction);
            };
            let Some(right_type) =
                infer_type_debrujin_int(gamma_free, gamma_bound.clone(), *rhs)?.into_concrete()
            else {
                return Err(InferError::ExpectedConreteType);
            };
            let out_type = match in_type.matches(&right_type) {
                (true, None) => out_type,
                (true, Some(a)) => out_type.specialize(&a, &right_type),
                (false, _) => return Err(InferError::MismatchedType),
            };
            Ok(out_type)
        }
        DeBrujinAst::FreeVariable(x) => gamma_free.get(&x).cloned().ok_or(InferError::NotInContext),
        // compiler bug if not present
        DeBrujinAst::BoundVariable(n) => Ok(gamma_bound.get(&n).cloned().unwrap()),
        DeBrujinAst::Constant(Constant::Nat(_)) => Ok(Type::Primitive(PrimitiveType::Nat).into()),
        DeBrujinAst::Constant(Constant::Bool(_)) => Ok(Type::Primitive(PrimitiveType::Bool).into()),
        DeBrujinAst::Constant(Constant::Float(_)) => {
            Ok(Type::Primitive(PrimitiveType::Float).into())
        }
    }
}
